from pathlib import Path
from ursus.config import config
from ursus.context_processors import Context, EntryContextProcessor, EntryURI
from ursus.utils import is_raster_image, get_image_size, get_image_transforms, is_image, is_pdf


class ImageProcessor(EntryContextProcessor):
    def process_entry(self, context: Context, entry_uri: EntryURI, changed_files: set[Path] | None = None) -> None:
        if config.fast_rebuilds and changed_files and (config.content_path / entry_uri) not in changed_files:
            return

        abs_path = config.content_path / entry_uri
        if is_raster_image(abs_path):
            width, height = get_image_size(abs_path)
            context['entries'][entry_uri].update({
                'width': width,
                'height': height,
            })
        if is_image(abs_path) or is_pdf(abs_path):
            context['entries'][entry_uri].update({
                'transforms': list(get_image_transforms(Path(entry_uri))),
            })



from . import Renderer
from pathlib import Path
from ursus.config import config
from ursus.context_processors import Context
from ursus.utils import get_files_in_path
import logging
import sass


logger = logging.getLogger(__name__)


class SassRenderer(Renderer):
    """
    Renders Sass .scss files as .css
    """

    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        files_to_keep = set()
        for scss_path in get_files_in_path(config.templates_path, changed_files, suffix='.scss'):
            output_path = scss_path.with_suffix('.css')

            logger.info('Rendering %s', str(output_path))
            with (config.output_path / output_path).open('w') as css_file:
                css_file.write(
                    sass.compile(
                        filename=str(config.templates_path / scss_path),
                        output_style='compressed' if config.minify_css else 'nested',
                        include_paths=[str(config.templates_path)],
                    )
                )
            files_to_keep.add(output_path)

        return files_to_keep



from dataclasses import dataclass, field
from markdown.extensions.wikilinks import build_url
from markdown.extensions.toc import slugify
from pathlib import Path
from platformdirs import user_cache_dir
from typing import Any, Callable, Iterable
import logging


def default_context_processors() -> list[str]:
    return [
        'ursus.context_processors.stale.StaleEntriesProcessor',
        'ursus.context_processors.image.ImageProcessor',
        'ursus.context_processors.markdown.MarkdownProcessor',
        'ursus.context_processors.get_entries.GetEntriesProcessor',
        'ursus.context_processors.related.RelatedEntriesProcessor',
        # 'ursus.context_processors.git_date.GitDateProcessor',
    ]


def default_image_transforms(max_size: int = 5000) -> dict:
    return {
        '': {
            'max_size': (max_size, max_size),
        },
    }


def default_jinja_filters() -> dict[str, Any]:
    return {}


def default_linters() -> list[str]:
    return [
        'ursus.linters.markdown.MarkdownLinkTextsLinter',
        'ursus.linters.markdown.MarkdownLinkTitlesLinter',
        'ursus.linters.markdown.MarkdownInternalLinksLinter',
        'ursus.linters.markdown.MarkdownExternalLinksLinter',
        'ursus.linters.markdown.RelatedEntriesLinter',
        'ursus.linters.images.UnusedImagesLinter',
    ]


def default_markdown_extensions() -> dict:
    return {
        'codehilite': {
            'guess_lang': False,
        },
        'fenced_code': {},
        'jinja': {},
        'meta': {},
        'responsive_images': {},
        'smarty': {},
        'superscript': {},
        'tables': {},
        'tasklist': {
            # The CSS class of Markdown list items with a checkbox ("- [ ] a list item")
            # Sets the class of the <li> and the <input type="checkbox">
            'list_item_class': None,
            'checkbox_class': None,
        },
        'toc': {
            'slugify': slugify,
        },
        'wikilinks': {
            # The base URL prepended to all markdown [[wikilinks]], without a trailing slash.
            # For example, https://allaboutberlin.com/glossary
            'base_url': '/',
            'end_url': '',
            'build_url': build_url,
            'html_class': None,
        },
        'better_footnotes': {
            'BACKLINK_TEXT': '⤴',
            'SUPERSCRIPT_TEXT': '{}',
        },
    }


def default_renderers() -> list[str]:
    return [
        'ursus.renderers.static.StaticAssetRenderer',
        'ursus.renderers.static.ArchiveRenderer',
        'ursus.renderers.image.ImageTransformRenderer',
        'ursus.renderers.jinja.JinjaRenderer',
        'ursus.renderers.lunr.LunrIndexRenderer',
        'ursus.renderers.sass.SassRenderer',
    ]


@dataclass
class UrsusConfig():
    content_path: Path = Path('content').resolve()
    templates_path: Path = Path('templates').resolve()
    output_path: Path = Path('output').resolve()
    cache_path: Path = Path(user_cache_dir('ursus', 'nicolasb'))

    # The URL of this website's root, without a trailing slash. For example, https://allaboutberlin.com
    site_url: str = ''

    # The URL extension of HTML pages. Change this if your server changes or removes the file extension.
    html_url_extension: str = '.html'

    # Minify Javascript and CSS
    minify_js: bool = False
    minify_css: bool = False

    # Builds the static website faster by only rebuilding templates for files that changed.
    # Related pages might not be rebuilt, even if they might be affected by the changes.
    # If False, everything is rebuilt from scratch. It's recommended to disable fast_rebuilds in production.
    fast_rebuilds: bool = False

    # Sets the <img sizes=""> attribute for your content images
    image_default_sizes: str | None = None

    # Transforms applied to your content images
    image_transforms: dict = field(default_factory=default_image_transforms)

    """
    Parametres to generate a search index for lunr.js
    Example:
    {
        'indexed_fields': ('body', ),  # Index these fields only
        'indexes': [
            {
                'uri_pattern': '*.md',  # Index entries with URIs that match this glob pattern
                'returned_fields': ('body', 'url', ),  # Return these fields only in the document list
                'boost': 1,  # Documents matching this pattern should be boosted n times over others
            },
        ]
    }
    """
    lunr_indexes: dict = field(default_factory=dict)
    lunr_index_output_path: Path = Path('search-index.json')  # Relative to output_path

    generator: str = 'ursus.generators.static.StaticSiteGenerator'

    # The processors that update the context with extra data
    context_processors: list[str] = field(default_factory=default_context_processors)
    context_globals: dict = field(default_factory=dict)

    markdown_extensions: dict = field(default_factory=default_markdown_extensions)

    # Translations
    translations_path: Path = Path('templates/_translations').resolve()
    default_language: str | None = None
    translation_languages: Iterable[str] | None = None

    openai_api_key: str | None = None
    metadata_fields_to_translate: Iterable[str] = ()

    # The renderers that take your templates and content, and populate the output dir
    renderers: list[str] = field(default_factory=default_renderers)

    # Linters look for errors in your content
    linters: list[str] = field(default_factory=default_linters)

    # Filter functions available in Jinja templates. The key is the filter name, and the value is a function
    jinja_filters: dict[str, Callable] = field(default_factory=default_jinja_filters)

    logging = {
        'datefmt': '%Y-%m-%d %H:%M:%S',
        'format': '%(asctime)s %(levelname)s [%(name)s:%(lineno)d] %(message)s',
        'level': logging.INFO,
    }

    def add_markdown_extension(self, name: str, config: dict[str, Any] = {}) -> None:
        self.markdown_extensions[name] = config


config = UrsusConfig()



from pathlib import Path
from typing import Generator, Match
from ursus.config import config
import logging
import re


logger = logging.getLogger(__name__)


LinterResult = Generator[tuple[tuple[int, int, int] | None, str, int], None, None]
LineLinterResult = Generator[tuple[tuple[int, int] | None, str, int], None, None]
MatchResult = Generator[tuple[str, int], None, None]


class Linter():
    def lint(self, file_path: Path) -> LinterResult:
        """
        Lints the content for errors.
        """
        raise NotImplementedError


class LineLinter(Linter):
    """
    Lints a text file line by line.
    """
    file_suffixes: tuple[str, ...] = ()

    def lint(self, file_path: Path) -> LinterResult:
        if self.file_suffixes and file_path.suffix.lower() not in self.file_suffixes:
            return

        with (config.content_path / file_path).open() as file:
            for line_no, line in enumerate(file.readlines()):
                for col_range, error, level in self.lint_line(file_path, line):
                    yield (line_no, *col_range), error, level

    def lint_line(self, file_path: Path, line: str) -> LineLinterResult:
        raise NotImplementedError


class RegexLinter(LineLinter):
    regex = re.compile(r'')

    def lint_line(self, file_path: Path, line: str) -> LineLinterResult:
        for match in self.regex.finditer(line):
            for message, level in self.handle_match(file_path, match):
                yield match.span(), message, level

    def handle_match(self, file_path: Path, match: Match[str]) -> MatchResult:
        raise NotImplementedError



from ursus.config import config
from ursus.utils import import_class, get_files_in_path
from . import Generator
import logging


logger = logging.getLogger(__name__)


class StaticSiteGenerator(Generator):
    """
    Turns a group of files and templates into a static website
    """

    def __init__(self):
        super().__init__()

        self.context_processors = [
            import_class(class_name)()
            for class_name in config.context_processors
        ]

        self.renderers = [
            import_class(class_name)()
            for class_name in config.renderers
        ]

        self.context = {
            **config.context_globals,
            'config': config,
            'entries': {},
        }

    def get_watched_paths(self):
        return [*super().get_watched_paths(), config.templates_path]

    def generate(self, changed_files=None):
        """
        Build a rendering context from the content
        """
        if config.fast_rebuilds and changed_files is not None:
            logger.info("Updating context...")
        else:
            logger.info("Building context...")

        for file_path in get_files_in_path(config.content_path, changed_files):
            entry_uri = str(file_path)
            self.context['entries'].setdefault(entry_uri, {
                'entry_uri': entry_uri,
            })

        for context_processor in self.context_processors:
            self.context = context_processor.process(self.context, changed_files)

        """
        Render entries and other templates
        """
        files_to_keep = set()
        for renderer in self.renderers:
            logger.debug(f"Rendering entries with {type(renderer).__name__}")
            files_to_keep.update(renderer.render(self.context, changed_files))

        """
        Delete output files older than this build. This is how stale output files are deleted.
        """
        if not config.fast_rebuilds:
            for file in config.output_path.rglob('*'):
                if file.is_file() and file.relative_to(config.output_path) not in files_to_keep:
                    logger.warning(f"Deleting stale output file {str(file.relative_to(config.output_path))}")
                    file.unlink()

        logger.info("Done.")



from collections import UserDict
from pathlib import Path
from typing import Any, NewType


EntryURI = NewType('EntryURI', str)
Entry = NewType('Entry', dict[str, Any])


class Context(UserDict[str, Any]):
    pass


class ContextProcessor:
    def process(self, context: Context, changed_files: set[Path] | None = None) -> Context:
        """Transforms the context and returns it. The context is used to render templates.

        Args:
            context (dict): An object that represents all data used to render templates
                (website info, blog posts, utility functions, etc.)
            changed_files (set, optional): A list of files that changed since the last context update.

        Returns:
            dict: The updated context
        """
        return context


class EntryContextProcessor(ContextProcessor):
    def process(self, context: Context, changed_files: set[Path] | None = None) -> Context:
        for entry_uri in list(context['entries'].keys()):
            self.process_entry(context, entry_uri, changed_files)
        return context

    def process_entry(self, context: Context, entry_uri: EntryURI, changed_files: set[Path] | None = None) -> None:
        raise NotImplementedError



from . import Renderer
from pathlib import Path
from PIL import Image
from ursus.config import config
from ursus.context_processors import Context
from ursus.utils import make_image_thumbnail, make_pdf_thumbnail, is_pdf, is_svg, copy_file
import logging


logger = logging.getLogger(__name__)


class ImageTransformRenderer(Renderer):
    """
    Resizes images and generate PDF thumbnails
    """

    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        logger.info("Rendering image transforms...")

        files_to_keep = set()

        for entry_uri, entry in context['entries'].items():
            abs_file_path = config.content_path / entry_uri

            for transform in entry.get('transforms', []):
                output_path = transform['output_path']
                abs_output_path = config.output_path / output_path
                max_size = transform.get('max_size')  # Optional for PDFs and SVGs

                has_changed = changed_files is None or abs_file_path in changed_files
                if has_changed and (
                    (not abs_output_path.exists())
                    or abs_file_path.stat().st_mtime > abs_output_path.stat().st_mtime
                ):
                    if is_pdf(abs_file_path):
                        if abs_output_path.suffix.lower() == '.pdf':
                            logger.info('Copying %s to %s', entry_uri, str(output_path))
                            copy_file(abs_file_path, abs_output_path)
                        else:
                            logger.info('Generating %s preview as %s', entry_uri, str(output_path))
                            make_pdf_thumbnail(abs_file_path, max_size, abs_output_path)
                    elif is_svg(abs_file_path):
                        logger.info('Copying %s to %s', entry_uri, str(output_path))
                        copy_file(abs_file_path, abs_output_path)
                    else:
                        logger.info('Converting %s to %s', entry_uri, str(output_path))
                        with Image.open(abs_file_path) as pil_image:
                            make_image_thumbnail(pil_image, max_size, abs_output_path)

                files_to_keep.add(output_path)

        return files_to_keep



from importlib.resources import files
from subprocess import run, STDOUT
from ursus.utils import import_class, get_files_in_path, log_color, log_color_end
from ursus.config import config
from watchdog.observers import Observer
import logging
import sys
import time


def build(watch_for_changes: bool = False) -> None:
    """Runs ursus and builds a static website

    Args:
        watch_for_changes (bool, optional): Keep running, and rebuild when content or templates change
    """
    generator = import_class(config.generator)()

    if watch_for_changes:
        observer = Observer()

        try:
            generator.generate()
        except:
            logging.exception("Could not generate site")

        for path in generator.get_watched_paths():
            observer.schedule(generator.get_observer_event_handler(), path, recursive=True)
        observer.start()
        try:
            while True:
                time.sleep(1)
        finally:
            observer.stop()
            observer.join()
    else:
        generator.generate()


def lint(files_to_lint=None, min_level=logging.INFO) -> None:
    """Lints the content for errors"""
    linters = [import_class(linter_path)() for linter_path in config.linters]

    has_errors = False

    if files_to_lint:
        logging.info(f"Linting {', '.join(map(str, files_to_lint))}")

    for file_path in sorted(get_files_in_path(config.content_path, whitelist=files_to_lint)):
        for linter in linters:
            linter_errors = list(linter.lint(file_path))
            for position, message, level in linter_errors:
                if position:
                    line_no, col_start, col_end = position
                else:
                    line_no = 0
                    col_start = 0
                    col_end = 1

                if level >= min_level:
                    has_errors = True
                    if line_no is not None:
                        logging.log(level, f"{log_color(level)}{str(file_path)}:{line_no}:{col_start}-{col_end}{log_color_end()} - {message}")
                    else:
                        logging.log(level, f"{str(file_path)} - {message}")
    sys.exit(1 if has_errors else 0)


def translate() -> None:
    babel_config = files("ursus") / 'babel' / 'pybabel.cfg'
    pot_path = config.translations_path / 'messages.pot'

    if not config.default_language or not config.translation_languages:
        raise Exception("Translations are not configured. You must set config.default_language and config.translation_languages.")

    config.translations_path.mkdir(parents=True, exist_ok=True)
    run(
        [
            'pybabel', 'extract',
            '--ignore-dirs', '.*',  # Prevent files/dirs starting with an underscore from being ignored
            '--mapping', str(babel_config),
            '--output-file', str(pot_path), str(config.templates_path)
        ],
        check=True, stdout=sys.stdout, stderr=STDOUT
    )
    for language_code in set([config.default_language, *config.translation_languages]):
        command = 'update' if (config.translations_path / language_code / 'LC_MESSAGES' / 'messages.po').exists() else 'init'
        run(
            [
                'pybabel', command,
                '--input-file', str(pot_path),
                '--locale', language_code,
                '--output-dir', str(config.translations_path),
            ],
            check=True, stdout=sys.stdout, stderr=STDOUT
        )
    run(
        ['pybabel', 'compile', '--directory', str(config.translations_path)],
        check=True, stdout=sys.stdout, stderr=STDOUT
    )



from pathlib import Path
from ursus.config import config
from watchdog.events import FileSystemEventHandler
import logging
import threading


logger = logging.getLogger(__name__)


class GeneratorObserverEventHandler(FileSystemEventHandler):
    def __init__(self, generator: 'Generator', **kwargs):
        self.generator = generator

        self.queued_events: set = set()
        self.debounce_timer: threading.Timer | None = None

        self.is_rebuilding: bool = False

        return super().__init__(**kwargs)

    def reschedule_rebuild(self) -> None:
        if self.debounce_timer:
            self.debounce_timer.cancel()
        self.debounce_timer = threading.Timer(0.5, self.on_file_changes)
        self.debounce_timer.start()

    def dispatch(self, event) -> None:
        if event.event_type in ('created', 'modified', 'moved', 'deleted'):
            self.queued_events.add(event)
            self.reschedule_rebuild()

    def on_file_changes(self) -> None:
        if self.is_rebuilding:
            self.reschedule_rebuild()
            return

        self.is_rebuilding = True

        changed_files = set(Path(event.src_path) for event in self.queued_events)
        changed_files.update([Path(e.dest_path) for e in self.queued_events if e.event_type == 'moved'])
        self.queued_events.clear()
        if len(changed_files):
            try:
                self.generator.on_file_changes(changed_files)
            except:
                logging.exception("Could not generate site")
        self.is_rebuilding = False


class Generator:
    def generate(self, changed_files: set[Path] | None = None) -> None:
        raise NotImplementedError

    def get_watched_paths(self) -> list[Path]:
        return [config.content_path, ]

    def get_observer_event_handler(self) -> FileSystemEventHandler:
        return GeneratorObserverEventHandler(generator=self)

    def on_file_changes(self, changed_files: set) -> None:
        self.generate(changed_files=changed_files)



from . import Renderer
from pathlib import Path
from ursus.config import config
from ursus.context_processors import Context
from ursus.utils import get_files_in_path, copy_file
import logging


logger = logging.getLogger(__name__)


class StaticFileRenderer(Renderer):
    """
    Copies static files to `output_path`
    """

    def get_files_to_copy(self, changed_files: set[Path] | None = None) -> list[tuple[Path, Path]]:
        # Return list of tuples: (absolute_original_path, destination_relative_to_output_path)
        return []

    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        files_to_keep = set()
        for asset_path, rel_output_path in self.get_files_to_copy():
            abs_output_path = config.output_path / rel_output_path

            if changed_files is None or asset_path in changed_files:
                logger.info('Copying asset %s', str(rel_output_path))
                copy_file(asset_path, abs_output_path)
            files_to_keep.add(rel_output_path)

        return files_to_keep


class StaticAssetRenderer(StaticFileRenderer):
    """
    Copies static assets in `templates_path` to `output_path`
    """
    ignored_suffixes = ('.jinja', )

    def get_files_to_copy(self, changed_files: set[Path] | None = None) -> list[tuple[Path, Path]]:
        return [
            (config.templates_path / f, f)
            for f in get_files_in_path(config.templates_path)
            if f.suffix.lower() not in self.ignored_suffixes
        ]


class ArchiveRenderer(Renderer):
    """
    Copies archives in `content_path` to `output_path`
    """
    included_suffixes: tuple[str, ...] = ('.zip', '.rar', '.gz', '.7z')

    def get_assets_to_copy(self, changed_files: set[Path] | None = None) -> list[Path]:
        return [
            f for f in get_files_in_path(config.content_path)
            if f.suffix.lower() in self.included_suffixes
        ]

    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        files_to_keep = set()
        for asset_path in self.get_assets_to_copy():
            abs_output_path = config.output_path / asset_path

            if changed_files is None or config.content_path / asset_path in changed_files:
                logger.info('Copying static entry %s', str(asset_path))
                copy_file(config.content_path / asset_path, abs_output_path)
            files_to_keep.add(asset_path)

        return files_to_keep



from itertools import chain
from markdown.blockprocessors import HashHeaderProcessor
from pathlib import Path
from requests.exceptions import ConnectionError
from typing import Any, Match, List, Tuple
from urllib.parse import unquote, urlparse
from ursus.config import config
from ursus.linters import Linter, LinterResult, MatchResult, RegexLinter
from ursus.utils import parse_markdown_head_matter
import logging
import re
import requests


class MarkdownLinksLinter(RegexLinter):
    """
    Verifies external links in Markdown files
    """
    file_suffixes = ('.md', )

    # Matches [], supports escaped brackets, ignores ![] images.
    first_half = r"(?P<first_half>!?\[(?P<alt_text>([^\]]|\\\])*)(?<!\\)\])"

    # Matches (), supports escaped parentheses
    second_half = r"(?P<second_half>\((?P<url>([^\) ]|\\\))*)(\s+(?P<caption>\".*\"))?(?<!\\)\))"

    regex = re.compile(first_half + second_half)

    def handle_match(self, file_path: Path, match: Match[str]) -> MatchResult:
        alt_text = match['alt_text'].strip()
        is_image = match['first_half'].startswith('!')

        caption = match['caption']

        for error, level in chain(
            self.validate_link_alt_text(alt_text, is_image, file_path),
            self.validate_link_caption(caption, is_image, file_path),
            self.validate_link_url(match['url'], is_image, file_path),
        ):
            yield f"{error}: {match.group(0)}", level

    def validate_link_alt_text(self, text: str, is_image: bool, file_path: Path):
        return
        yield

    def validate_link_url(self, url: str, is_image: bool, file_path: Path):
        return
        yield

    def validate_link_caption(self, caption: str, is_image: bool, file_path: Path):
        return
        yield


class MarkdownLinkTextsLinter(MarkdownLinksLinter):
    def validate_link_alt_text(self, text: str, is_image: bool, file_path: Path):
        if not text:
            yield "Image has no alt text", logging.WARNING


class MarkdownLinkTitlesLinter(MarkdownLinksLinter):
    def validate_link_caption(self, caption: str, is_image: bool, file_path: Path):
        if caption is not None:
            if not (caption.startswith('"') and caption.endswith('"')):
                yield "Image caption is not quoted", logging.ERROR
            elif caption == caption.strip('"'):
                yield "Image caption is empty", logging.WARNING


class MarkdownExternalLinksLinter(MarkdownLinksLinter):
    user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.response_cache = {}

    def validate_link_url(self, url: str, is_image: bool, file_path: Path):
        if not url:
            yield "Missing URL", logging.ERROR
        elif not url.startswith(('/', '#', 'http://', 'https://', 'mailto:', 'tel:')):
            yield "Relative or invalid URL", logging.WARNING
        else:
            if url.startswith(('http://', 'https://')):
                try:
                    cleaned_url = self.unescape_url(url)
                    if cleaned_url not in self.response_cache:
                        self.response_cache[cleaned_url] = requests.get(cleaned_url, timeout=5, headers={
                            'User-Agent': self.user_agent
                        })
                    response = self.response_cache[cleaned_url]
                    status_code = response.status_code
                except ConnectionError:
                    yield f"Connection error: {cleaned_url}", logging.ERROR
                except requests.exceptions.RequestException as exc:
                    yield f"URL {type(exc).__name__}: {cleaned_url}", logging.ERROR
                else:
                    if status_code in (404, 410):
                        yield f"URL returns HTTP {status_code}", logging.ERROR
                    elif status_code >= 400:
                        level = logging.WARNING if status_code in (403, 503) else logging.ERROR
                        yield f"URL returns HTTP {status_code}", level
                    elif response.history and response.history[-1].status_code == 301:
                        yield f"URL redirects to {response.url}", logging.INFO

    def escape_url(self, url: str) -> str:
        return url.replace('(', '\\(').replace(')', '\\)')

    def unescape_url(self, url: str) -> str:
        return url.replace('\\(', '(').replace('\\)', ')')


class MarkdownInternalLinksLinter(MarkdownLinksLinter):
    """
    Verify that internal links point to existing entries. If the URL has a fragment,
    it should point to an existing title fragment.
    """
    header_regex = HashHeaderProcessor.RE
    ignored_urls = (
        # re.compile(r'^/ignored-url$')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title_slugs_cache = {}

    def get_title_slugs(self, file_path: Path):
        if file_path not in self.title_slugs_cache:
            self.title_slugs_cache[file_path] = set()
            with file_path.open() as file:
                for line in file.readlines():
                    if bool(self.header_regex.search(line)):
                        self.title_slugs_cache[file_path].add(
                            config.markdown_extensions['toc']['slugify'](line.lstrip('#').strip(), '-')
                        )
        return self.title_slugs_cache[file_path]

    def validate_link_url(self, url: str, is_image: bool, current_file_path: Path):
        for ignored_url in self.ignored_urls:
            if ignored_url.match(url):
                return

        url_parts = urlparse(url)
        title_slug = url_parts.fragment

        # Convert relative and absolute URLs to content paths
        if config.site_url and url.startswith(config.site_url):
            file_path = config.content_path / unquote(url_parts.path.removeprefix(config.site_url))
        elif url.startswith('/'):
            file_path = config.content_path / unquote(url_parts.path.lstrip('/'))
        elif url.startswith('#'):
            file_path = config.content_path / current_file_path
        elif not url_parts.scheme and not url_parts.netloc:  # Relative URL
            file_path = current_file_path.parent / unquote(url_parts.path)
        else:
            return

        if file_path.suffix.lower() == config.html_url_extension:
            file_path = file_path.with_suffix('.md')

        if not file_path.exists():
            yield "Entry not found", logging.ERROR
        elif title_slug and title_slug not in self.get_title_slugs(file_path):
            yield "URL fragment not found", logging.ERROR


class HeadMatterLinter(Linter):
    def lint(self, file_path: Path) -> LinterResult:
        if file_path.suffix.lower() != '.md':
            return

        meta: dict[str, List[Any]] = {}
        field_positions: dict[str, Tuple[int, int, int]] = {}

        with (config.content_path / file_path).open() as file:
            meta, field_positions = parse_markdown_head_matter(file.readlines())

        yield from self.lint_meta(file_path, meta, field_positions)

    def lint_meta(self, file_path: Path, meta: dict[str, List[Any]], field_positions: dict[str, Tuple[int, int, int]]) -> LinterResult:
        raise NotImplementedError


class RelatedEntriesLinter(HeadMatterLinter):
    def lint_meta(self, file_path: Path, meta: dict[str, List[Any]], field_positions: dict[str, Tuple[int, int, int]]) -> LinterResult:
        for key in meta.keys():
            if key.startswith('related_'):
                for pos, entry_uri in enumerate(meta[key]):
                    if not (config.content_path / entry_uri).exists():
                        line_no, col, end_col = field_positions[key]
                        line_no += pos
                        yield (
                            (line_no, 4, 4 + len(entry_uri)),
                            f"Entry does not exist: {entry_uri}",
                            logging.ERROR
                        )



from . import Context, ContextProcessor, Entry, EntryURI
from functools import partial
from operator import itemgetter
from pathlib import Path
from typing import Any, Callable


def first_existing_item_getter(keys: list[str]) -> Any:
    """
    Returns the value of the first existing key from a list of dictionary keys

    Args:
        keys (list[str]): A list of dictionary keys
    """

    def get_value(entry: Entry) -> Any:
        return list(
            filter(None, [entry.get(key) for key in keys])
        )[0]
    return get_value


def get_entries(
    entries: dict[EntryURI, Entry],
    namespaces: str | list[str] | None = None,
    filter_by: Callable[[EntryURI, Entry], bool] | None = None,
    sort_by: Callable[[Entry], Any] | str | list[str] | None = None,
    reverse: bool = False
) -> list[Entry]:
    """Returns a sorted, filtered list of entries

    Args:
        entries: The dictionary of entries.
        namespace: Only returns entries in the given namespace(s) (for example "posts" or "blog/posts"). In other
            words, only return entries in a given directory (like <content_path>/posts or <content_path>/blog/posts).
        filter_by: Filter the items by the given filtering function.
        sort_by: Sort items by the given dict key, list of dict keys, or value
            returned by the given function
        reverse: Reverse the sorting order
    """
    if namespaces:
        namespace_list = [namespaces, ] if isinstance(namespaces, str) else namespaces
        entries = {
            uri: value for uri, value in entries.items()
            if uri.startswith(tuple(ns + '/' for ns in namespace_list))
        }

    if filter_by:
        entries = {
            uri: value for uri, value in entries.items()
            if filter_by(uri, value)
        }

    entry_list = list(entries.values())

    if sort_by:
        if callable(sort_by):
            sorter = sort_by
        elif isinstance(sort_by, str):
            sorter = itemgetter(sort_by)
        else:
            sorter = first_existing_item_getter(sort_by)
        entry_list = sorted(entry_list, key=sorter, reverse=reverse)

    return entry_list


class GetEntriesProcessor(ContextProcessor):
    """
    Adds the get_entries() method to the context root. This function filters and
    sorts entries.
    """

    def process(self, context: Context, changed_files: set[Path] | None = None) -> Context:
        if 'get_entries' not in context:
            context['get_entries'] = partial(get_entries, context['entries'])
        return context



from ursus.context_processors import Context
from pathlib import Path


class Renderer:
    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        """Creates, updates or touches files in output_path

        Args:
            context (dict): Context used to render this file
            changed_files (set, optional): A list of changed content and
                template files (absolute paths)
        Returns:
            set: List of output files that should be preserved. Files that are
                not preserved by any renderer will marked as stale and deleted.

        Raises:
            NotImplementedError: Description
        """
        raise NotImplementedError



from pathlib import Path
from ursus.config import config
from ursus.context_processors import Context, ContextProcessor


class StaleEntriesProcessor(ContextProcessor):
    """
    Removes entries if their content file no longer exists
    """

    def process(self, context: Context, changed_files: set[Path] | None = None) -> Context:
        for file in (changed_files or set()):
            if file.is_relative_to(config.content_path) and not file.exists():
                entry_uri = str(file.relative_to(config.content_path))
                try:
                    context['entries'].pop(entry_uri)
                except KeyError:
                    pass

        return context



from . import Renderer
from jinja2 import Environment, FileSystemLoader, nodes, pass_context, select_autoescape, StrictUndefined
from jinja2.ext import Extension, do
from jinja2.meta import find_referenced_templates
from jinja2_simple_tags import StandaloneTag, ContainerTag
from markdown.serializers import to_html_string
from markupsafe import Markup
from ordered_set import OrderedSet
from pathlib import Path
from rjsmin import jsmin
from rcssmin import cssmin
from typing import Generator
from ursus.config import config
from ursus.context_processors import Context, EntryURI
from ursus.utils import get_files_in_path, make_picture_element, is_ignored_file
import gettext
import logging
import sass


logger = logging.getLogger(__name__)


class MultilingualGNUTranslations(gettext.GNUTranslations):
    """
    Loads translations in different languages based on the "language" variable in the template context.
    """
    translations: dict[str, gettext.GNUTranslations] = {}

    def __init__(self, *args, **kwargs):
        for language_code in set([config.default_language, *config.translation_languages]):
            self.translations[language_code] = gettext.translation(
                domain='messages',
                localedir=config.translations_path,
                languages=[language_code],
                fallback=True,
            )

    @pass_context
    def gettext(self, context, *args, **kwargs):
        logging.info(f"Translating to {context.get('language', config.default_language)}")
        return self.translations[context.get('language', config.default_language)].gettext(*args, **kwargs)


class JsLoaderExtension(Extension):
    """
    Jinja extension. Adds the {% js %} and {% alljs %} tags.

    All javascript code between {% js %} tags is combined and queued for future
    output. It's minified if config.minify_js is True.

    {% alljs %} outputs the queued javascript code. The code is included in
    order of addition. If the same code is queued multiple times, it's only
    output once.
    """
    tags = {"js", "alljs"}

    def __init__(self, environment):
        super().__init__(environment)
        environment.extend(js_fragments=OrderedSet())

    def parse(self, parser):
        token = next(parser.stream)

        if token.test('name:alljs'):
            call = self.call_method('_render_js', args=[nodes.ContextReference()])
            return nodes.Output([nodes.MarkSafe(call)]).set_lineno(token.lineno)
        else:
            body = parser.parse_statements(["name:endjs"], drop_needle=True)
            return nodes.CallBlock(
                self.call_method("_queue_js"), [], [], body
            ).set_lineno(token.lineno)

    def _render_js(self, caller):
        output = "\n".join(self.environment.js_fragments)
        if config.minify_js:
            output = jsmin(output)
        self.environment.js_fragments.clear()
        return Markup(output)

    def _queue_js(self, caller):
        self.environment.js_fragments.add(caller())
        return ''


class CssLoaderExtension(ContainerTag):
    """
    Jinja extension. Adds the {% css %} tag.

    All CSS code in {% css %} tags is minified if config.minify_css is True.
    """
    tags = {"css"}
    safe_output = True

    def render(self, caller):
        output = caller()
        if config.minify_css:
            output = cssmin(output)
        return Markup(output)


class ScssLoaderExtension(ContainerTag):
    """
    Jinja extension. Adds the {% scss %} tag.

    All Sass code in {% scss %} tags is converted to CSS.
    It's also minified if config.minify_css is True.
    """
    tags = {"scss"}
    safe_output = True

    def __init__(self, *args, **kwargs):
        self.scss_cache = {}

    def render(self, caller):
        scss_code = caller()
        if scss_code not in self.scss_cache:
            self.scss_cache[scss_code] = Markup(
                sass.compile(
                    string=scss_code,
                    output_style='compressed' if config.minify_css else 'nested',
                    include_paths=[str(config.templates_path)],
                )
            )
        return self.scss_cache[scss_code]


class ResponsiveImageExtension(StandaloneTag):
    """Jinja extension. Adds {% image %} tag.

    Usage: {% image 'relative/path/to/original.jpg' img_class='a_css_class' sizes='(min-width: 600px) 160px, 320px' %}

    This tag is replaced by a responsive <picture> element. The correct image transforms for the given source image are
    used.
    """
    tags = {"image"}
    safe_output = True

    def render(self, image_entry_uri, css_class=None, alt=None, sizes=None):
        img_attrs = {}

        if css_class:
            img_attrs['class'] = css_class
        if alt:
            img_attrs['alt'] = alt

        # Render the same way as Python Markdown does, for consistency
        return to_html_string(make_picture_element(self.context, image_entry_uri, img_attrs, sizes))


@pass_context
def render_filter(context, value) -> str:
    return context.eval_ctx.environment.from_string(value).render(**context)


class JinjaRenderer(Renderer):
    """
    Renders all .jinja templates in the templates directory, unless their name starts with '_'.
    """

    def __init__(self):
        super().__init__()

        translations = gettext.translation(
            domain='messages',
            localedir=config.translations_path,
            languages=['de'],
            fallback=True,
            class_=MultilingualGNUTranslations
        )

        self.template_environment = Environment(
            loader=FileSystemLoader(config.templates_path),
            extensions=['jinja2.ext.i18n', do, JsLoaderExtension, CssLoaderExtension, ScssLoaderExtension, ResponsiveImageExtension],
            autoescape=select_autoescape(),
            undefined=StrictUndefined
        )

        self.template_environment.install_gettext_translations(translations, newstyle=True)
        self.template_environment.filters['render'] = render_filter
        self.template_environment.filters.update(config.jinja_filters)

    def get_child_templates(self, template_path: Path) -> set[Path]:
        dependencies = set()
        with (config.templates_path / template_path).open() as template_file:
            ast = self.template_environment.parse(template_file.read())
        child_template_paths = [Path(t.removeprefix('/')) for t in find_referenced_templates(ast)]
        for child_template_path in child_template_paths:
            dependencies.add(child_template_path)
            dependencies.update(self.get_child_templates(child_template_path))
        return dependencies

    def is_entry_template(self, template_path: Path) -> bool:
        return template_path.with_suffix('').stem == 'entry'

    def template_can_render_entry(self, template_path: Path, context: Context, entry_uri: EntryURI) -> bool:
        entry_path = Path(entry_uri)
        if entry_path.parent != template_path.parent:
            return False

        is_dedicated_template_for_this_entry = template_path.with_suffix('').stem == entry_path.stem
        if is_dedicated_template_for_this_entry:
            return True

        template_suffixes = "".join([template_path.with_suffix('').suffix, template_path.suffix])
        entry_has_dedicated_template = (
            config.templates_path / template_path.parent / (entry_path.stem + template_suffixes)
        ).exists()
        if self.is_entry_template(template_path) and not entry_has_dedicated_template:
            return True

        return False

    def render_template(self, template_path: Path, context: Context, output_path: Path) -> Generator[Path, None, None]:
        """Returns an entry into a template, and saves it under output_path
        Args:
            template_path (Path): Path to the template to use, relative to output_path.
            context (dict): Context object used to render the template.
            output_path (str): Path of the generated file, relative to the output_path.
        """
        logger.info('Rendering %s', str(output_path))
        abs_output_path = config.output_path / output_path
        abs_output_path.parent.mkdir(parents=True, exist_ok=True)
        template = self.template_environment.get_template(str(template_path))
        template.stream(**context).dump(str(abs_output_path))
        yield output_path

    def get_entry_output_path(self, template_path: Path, entry_uri: EntryURI) -> Path:
        """Gets the path where the rendered template will be saved
        Args:
            template_path (Path): Template used to render the entry
            entry_uri (str): URI of the entry to render
        Returns:
            Path: Rendered template output path, relative to the output_path.
        """
        output_suffix = template_path.with_suffix('').suffix  # Remove .jinja, so that .html.jinja becomes .html
        return Path(entry_uri).with_suffix(output_suffix)

    def render_entry(self, template_path: Path, context: Context, entry_uri: EntryURI) -> Generator[Path, None, None]:
        """Returns an entry into a template, and saves it under output_path
        Args:
            template_path (Path): Path to the template to use, relative to output_path.
            context (dict): Context object used to render the template.
            entry_uri (str): URI of the entry to render
        Returns:
            Path: Path of the generated file, relative to the output_path.
        """
        specific_context = Context({
            **context,
            'entry': context['entries'][entry_uri],
            'entry_uri': entry_uri,
        })
        output_path = self.get_entry_output_path(template_path, entry_uri)
        yield from self.render_template(template_path, specific_context, output_path)

    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        template_paths = get_files_in_path(config.templates_path, suffix='.jinja')

        render_queue: OrderedSet = OrderedSet()
        files_to_keep: set[Path] = set()

        changed_entry_uris = set()
        changed_templates = set()
        for file in (changed_files or set()):
            if not (file.exists() and file.is_file()):
                continue

            if file.is_relative_to(config.content_path) and not is_ignored_file(file, config.content_path):
                changed_entry_uris.add(EntryURI(str(file.relative_to(config.content_path))))
            elif file.is_relative_to(config.templates_path):
                changed_templates.add(file.relative_to(config.templates_path))
            else:
                continue

        if changed_files is not None and config.fast_rebuilds:
            # Also rerender templates that depend on the changed templates (_style.css > _layout.html > index.html)
            changed_parent_templates = set()
            for template_path in template_paths:
                dependencies = self.get_child_templates(template_path)
                for changed_template in changed_templates:
                    if changed_template in dependencies:
                        logger.info(f"{template_path} is affected by {changed_template} change")
                        changed_parent_templates.add(template_path)
            changed_templates.update(changed_parent_templates)

        # Process edited entries
        for entry_uri in changed_entry_uris:
            for tp in template_paths:
                if self.template_can_render_entry(tp, context, entry_uri):
                    render_queue.add(('entry', tp, entry_uri))

        # Process edited templates
        for template_path in changed_templates:
            if is_ignored_file(config.templates_path / template_path, config.templates_path):
                continue

            can_render_an_entry = False
            for entry_uri in context['entries']:
                if self.template_can_render_entry(template_path, context, entry_uri):
                    render_queue.add(('entry', template_path, entry_uri))
                    can_render_an_entry = True

            if not self.is_entry_template(template_path) and not can_render_an_entry:
                render_queue.add(('template', template_path, template_path.with_suffix('')))

        # Process everything else
        for template_path in template_paths:
            can_render_an_entry = False
            for entry_uri in context['entries']:
                if self.template_can_render_entry(template_path, context, entry_uri):
                    can_render_an_entry = True
                    if config.fast_rebuilds:
                        output_file = (config.output_path / self.get_entry_output_path(template_path, entry_uri))
                        output_file.parent.mkdir(exist_ok=True)
                        output_file.touch()
                    else:
                        render_queue.add(('entry', template_path, entry_uri))

            if not self.is_entry_template(template_path) and not can_render_an_entry:
                output_path = template_path.with_suffix('')  # Remove .jinja
                if config.fast_rebuilds:
                    output_file = (config.output_path / output_path)
                    output_file.parent.mkdir(exist_ok=True)
                    output_file.touch()
                else:
                    render_queue.add(('template', template_path, output_path))

        for render_type, template_path, value in render_queue:
            if render_type == 'entry':
                files_to_keep.update(self.render_entry(template_path, context, value))
            elif render_type == 'template':
                files_to_keep.update(self.render_template(template_path, context, value))

        return files_to_keep



from datetime import datetime
from pathlib import Path
from ursus.config import config
from ursus.context_processors import Context, ContextProcessor, EntryURI
import git
import logging


def unescape_backslashes(s: str, encoding: str = 'utf-8') -> str:
    """
    Convert backslash-escaped strings to normal strings
    ("DerHimmel\303\234berBerlin" -> "DerHimmelÜberBerlin")
    """
    return (s.encode('latin1')
             .decode('unicode-escape')
             .encode('latin1')
             .decode(encoding))


class GitDateProcessor(ContextProcessor):
    """
    Sets entry.date_updated to the date of the latest commit.
    """

    def __init__(self: 'GitDateProcessor'):
        super().__init__()
        self.entry_uri_commit_dates: dict[EntryURI, datetime] = {}
        if not config.fast_rebuilds:
            self.repo = git.Repo(config.content_path, search_parent_directories=True)
            self.repo_root = Path(self.repo.working_dir)

            commit_date = None
            git_log = self.repo.git.log('--format=">>>%cd"', '--date=unix', '--name-only', '--encoding=UTF-8')
            for line in git_log.split('\n'):
                # Remove wrapping quotes, convert backslash-escaped unicode characters back to unicode
                line = unescape_backslashes(line.strip('"'))
                if line.startswith('>>>'):
                    commit_date = datetime.fromtimestamp(int(line.removeprefix('>>>'))).astimezone()
                elif len(line) > 0:
                    entry_uri = self.commit_path_to_entry_uri(line)
                    if entry_uri in self.entry_uri_commit_dates:
                        self.entry_uri_commit_dates[entry_uri] = max(commit_date, self.entry_uri_commit_dates[entry_uri])
                    else:
                        self.entry_uri_commit_dates[entry_uri] = commit_date

    def commit_path_to_entry_uri(self, commit_path: str) -> EntryURI | None:
        abs_commit_path = self.repo_root / commit_path
        try:
            return EntryURI(str(abs_commit_path.relative_to(config.content_path)))
        except ValueError:
            return None

    def process(self, context: Context, changed_files: set[Path] | None = None) -> Context:
        for entry_uri, entry in context['entries'].items():
            if entry_uri in self.entry_uri_commit_dates:
                entry['date_updated'] = self.entry_uri_commit_dates.get(entry_uri)
            else:
                if not config.fast_rebuilds:
                    logging.warning(f"Entry {entry_uri} has no commit date")
                entry['date_updated'] = datetime.now().astimezone()
        return context



from http.server import SimpleHTTPRequestHandler
from threading import Thread
from ursus.config import config
import logging
import socketserver


class HttpRequestHandler(SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=config.output_path, **kwargs)

    def do_GET(self):
        abs_path = config.output_path / self.path.removeprefix('/')
        abs_html_path = abs_path.with_suffix('.html')
        abs_index_path = abs_path / 'index.html'

        if not abs_path.exists():
            if abs_path.suffix == config.html_url_extension and abs_html_path.exists():
                self.path = str(abs_html_path.relative_to(config.output_path))
            elif abs_index_path.exists():
                self.path = str(abs_index_path.relative_to(config.output_path))

        try:
            return super().do_GET()
        except BrokenPipeError:
            # These errors have no impact
            pass

    def log_message(self, format, *args):
        logging.debug(f"Request to {self.path}")


def serve(port: int = 80) -> None:
    """Start a static file server that serves Ursus on the given port.

    Args:
        port (int, optional): The port on which to serve the static site. Default is port 80.
    """
    socketserver.ThreadingTCPServer.allow_reuse_address = True
    with socketserver.ThreadingTCPServer(("", port), HttpRequestHandler) as server:
        logging.info(f"Serving static site on port {port}")
        server.serve_forever()


def serve_async(port: int = 80) -> Thread:
    thread = Thread(target=serve, args=(port, ), daemon=True)
    thread.start()
    return thread



from . import Context, ContextProcessor, Entry, EntryURI
from collections import UserDict
from pathlib import Path
from typing import Any
import sys


class RelatedEntryReferenceDict(UserDict[str, Any]):
    def __init__(self, entry: Entry, all_entries: dict[EntryURI, Entry]):
        self.all_entries = all_entries
        super().__init__(entry)

    def __getitem__(self, key: str) -> Any:
        if key.startswith('related_') and key in self.data:
            related_value: list[str] | str = self.data[key]
            try:
                if isinstance(related_value, str):  # Single URI string
                    return [self.all_entries[EntryURI(related_value)]]
                else:  # List of URI strings
                    return [self.all_entries[EntryURI(subvalue)] for subvalue in related_value]
            except KeyError:
                raise ValueError(f"{key} contains invalid value {sys.exc_info()[1]}")
        return super().__getitem__(key)


class RelatedEntriesProcessor(ContextProcessor):
    """
    Entry fields that start with related_* return a list of entries, instead of
    a list of entry URIs.
    """

    def process(self, context: Context, changed_files: set[Path] | None = None) -> Context:
        for uri, entry in context['entries'].items():
            if not isinstance(context['entries'][uri], RelatedEntryReferenceDict):
                context['entries'][uri] = RelatedEntryReferenceDict(entry, context['entries'])

        return context



from . import Context, EntryContextProcessor, EntryURI
from datetime import datetime
from markdown import Markdown
from markdown.extensions import Extension
from markdown.extensions.footnotes import FootnoteExtension, FN_BACKLINK_TEXT, NBSP_PLACEHOLDER
from markdown.inlinepatterns import SimpleTagPattern
from markdown.postprocessors import RawHtmlPostprocessor
from markdown.preprocessors import Preprocessor
from markdown.treeprocessors import Treeprocessor
from pathlib import Path
from typing import Any
from ursus.config import config
from ursus.utils import make_figure_element, make_picture_element
from xml.etree import ElementTree
from xml.etree.ElementTree import Element
import logging
import re


logger = logging.getLogger(__name__)


class TaskListProcessor(Treeprocessor):
    box_checked = '[x] '
    box_unchecked = '[ ] '

    def run(self, root: Element) -> Element:
        for li in root.iter(tag='li'):
            text = (li.text or "")

            if text.lower().startswith((self.box_checked, self.box_unchecked)):
                is_checked = text.lower().startswith(self.box_checked)

                checkbox = Element("input", {'type': 'checkbox'})
                if is_checked:
                    checkbox.attrib['checked'] = 'checked'
                if self.md.getConfig('checkbox_class'):
                    checkbox.attrib['class'] = self.md.getConfig('checkbox_class')

                checkbox.tail = text.removeprefix(self.box_checked if is_checked else self.box_unchecked)
                li.text = ''
                li.insert(0, checkbox)
                if self.md.getConfig('list_item_class'):
                    css_classes = set(li.attrib.get('class', '').split())
                    css_classes.update(self.md.getConfig('list_item_class').split())
                    li.attrib['class'] = " ".join(css_classes)

        return root


class TaskListExtension(Extension):
    """
    Adds github-flavored markdown todo lists:

    * [ ] Unchecked item
    * [x] Checked item
    """

    def __init__(self, **kwargs):
        self.config = {
            "list_item_class": ['', 'CSS class to add to the <li> element'],
            "checkbox_class": ['', 'CSS class to add to the checkbox <input>'],
        }
        super().__init__(**kwargs)

    def extendMarkdown(self, md):
        md.treeprocessors.register(TaskListProcessor(self), "gfm-tasklist", 100)


class JinjaPreprocessor(Preprocessor):
    """
    Ignore Jinja {{ ... }} and {% ... %} tags.
    """
    JINJA_RE = re.compile('({{([^}]+)}})|({%([^}]+)%})', re.MULTILINE | re.DOTALL)

    def run(self, lines: list[str]) -> list[str]:
        text = "\n".join(lines)

        def replace_match(match):
            return self.md.htmlStash.store(match[0])

        return re.sub(self.JINJA_RE, replace_match, text).split("\n")


class JinjaHtmlPostProcessor(RawHtmlPostprocessor):
    """
    Patch RawHtmlPostprocessor to recognize {% ... %} tags as block-level
    elements, and prevent them from being wrapped in a <p> tag.
    """
    JINJA_BLOCK_RE = re.compile('^{%([^}]+)%}$', re.MULTILINE | re.DOTALL)

    def isblocklevel(self, html: str) -> bool:
        m = self.JINJA_BLOCK_RE.match(html)
        if m:
            return True
        return super().isblocklevel(html)


class JinjaExtension(Extension):
    """
    Escape Jinja {% ... %} and {{ ... }} statements in Markdown files.
    """

    def extendMarkdown(self, md):
        md.postprocessors.deregister('raw_html')
        md.preprocessors.register(JinjaPreprocessor(md), 'jinja', 25)
        md.postprocessors.register(JinjaHtmlPostProcessor(md), 'raw_html', 30)


class ResponsiveImageProcessor(Treeprocessor):
    allowed_parents = (
        'a', 'p', 'pre', 'ul', 'ol', 'dl', 'div', 'blockquote', 'noscript', 'section', 'nav', 'article',
        'aside', 'header', 'footer', 'table', 'form', 'fieldset', 'menu', 'canvas', 'details'
    )

    def _swap_element(self, parent: Element, old: Element, new: Element) -> None:
        """
        Replaces `old` element with `new` in `parent` element
        """
        for index, element in enumerate(parent):
            if element == old:
                parent[index] = new
                new.tail = old.tail
                return

    def _upgrade_img(self, img: Element, parents: list[Element]) -> None:
        # Create <picture> with <source> for the different image types and sizes
        # Only apply to local images
        img_src = img.attrib.get('src', '')
        if img_src.startswith('/') or img_src.startswith(config.site_url + '/'):
            image_uri = EntryURI(img_src.removeprefix(config.site_url).removeprefix('/'))

            parent = parents[0]
            grandparent = parents[1]

            def has_single_child(element: Element) -> bool:
                return len(element) == 1 and not element.text

            img_attrs = img.attrib
            a_attrs = None
            image_maker = make_figure_element

            element_to_swap = img
            containing_element = parent

            # A valid <figure> parent with an empty <a> wrapping this <img>
            # In this case, wrap the <figure> around the <a>
            # li > a > img becomes li > figure > a > picture
            if (
                parent.tag == 'a'
                and has_single_child(parent)
                and grandparent.tag in self.allowed_parents
            ):
                a_attrs = parent.attrib
                element_to_swap = parent
                containing_element = grandparent

                # An empty <p> with an empty <a> with this <img>
                # Replace the whole thing with a <figure> containing the <a>
                # p > a > img becomes p > figure > a > picture
                if grandparent.tag == 'p' and has_single_child(grandparent):
                    element_to_swap = grandparent

                    greatgrandparent = parents[2]
                    containing_element = greatgrandparent

            # An empty <p> with this <img>
            # p > img becomes figure > picture
            elif parent.tag == 'p' and has_single_child(parent):
                element_to_swap = parent
                containing_element = grandparent

            # This element does not allow a <figure>. Just use a <picture>.
            elif parent.tag not in self.allowed_parents:
                image_maker = make_picture_element

            image = image_maker(self.md.context, image_uri, img_attrs, a_attrs, sizes=config.image_default_sizes)

            self._swap_element(containing_element, element_to_swap, image)

    def run(self, root: Element) -> None:
        parent_map = {}
        for parent in root.iter():
            for child in parent:
                parent_map[child] = parent

        for img in root.iter('img'):
            child = img
            parents = []
            while parent := parent_map.get(child):
                parents.append(parent)
                child = parent

            self._upgrade_img(img, parents)


class ResponsiveImagesExtension(Extension):
    """
    Transforms how <img> tags are rendered:

    - Adds srcset= to images to make them responsive
    - Wraps block images in a <figure> tag, and replaces the title with a <figcaption>

    """

    def extendMarkdown(self, md):
        md.registerExtension(self)
        self.md = md
        self.reset()
        md.treeprocessors.register(ResponsiveImageProcessor(md), 'figure', 0)

    def reset(self) -> None:
        pass


class SuperscriptExtension(Extension):
    """
    ^text^ is converted to <sup>text</sup>
    """

    # match ^, at least one character that is not ^, and ^ again
    SUPERSCRIPT_RE = r"(\^)([^\^]+)\2"

    def extendMarkdown(self, md):
        """Insert 'superscript' pattern before 'not_strong' pattern (priority 70)."""

        md.inlinePatterns.register(SimpleTagPattern(self.SUPERSCRIPT_RE, "sup"), 'superscript', 60)


class FootnotesExtension(FootnoteExtension):
    def extendMarkdown(self, md):
        super().extendMarkdown(md)

    def makeFootnotesDiv(self, root: Element) -> Element | None:
        if not list(self.footnotes.keys()):
            return None

        container = Element("details")
        container.set("class", "footnote")
        container.set("id", "footnotes")
        summary = ElementTree.SubElement(container, "summary")
        summary.text = "Sources and footnotes"
        ol = ElementTree.SubElement(container, "ol")

        surrogate_parent = Element("div")

        backlink_title = self.getConfig("BACKLINK_TITLE").replace("%d", "{}")

        for index, id in enumerate(self.footnotes.keys(), start=1):
            li = ElementTree.SubElement(ol, "li")
            li.set("id", self.makeFootnoteId(id))
            # Parse footnote with surrogate parent as li cannot be used.
            # List block handlers have special logic to deal with li.
            # When we are done parsing, we will copy everything over to li.
            self.parser.parseChunk(surrogate_parent, self.footnotes[id])
            for el in list(surrogate_parent):
                li.append(el)
                surrogate_parent.remove(el)
            backlink = Element("a")
            backlink.set("href", f"#{self.makeFootnoteRefId(id)}")
            backlink.set("class", "footnote-backref")
            backlink.set(
                "title",
                backlink_title.format(index)
            )
            backlink.text = FN_BACKLINK_TEXT

            if len(li):
                node = li[-1]
                if node.tag == "p":
                    node.text = node.text + NBSP_PLACEHOLDER
                    node.append(backlink)
                else:
                    p = ElementTree.SubElement(li, "p")
                    p.append(backlink)

        return container


class MarkdownProcessor(EntryContextProcessor):
    def __init__(self):
        super().__init__()

        self.markdown = Markdown(
            output_format='html',
            extensions=list(config.markdown_extensions.keys()),
            extension_configs=config.markdown_extensions,
        )

    def parse_metadata(self, raw_metadata: dict[str, Any]) -> dict[str, Any]:
        metadata = {}
        for key, value in raw_metadata.items():
            if len(value) == 0:
                continue
            if len(value) == 1:
                value = value[0]

            if key.startswith('date_'):
                value = datetime.strptime(value, '%Y-%m-%d').astimezone()

            if key.startswith('related_'):
                if isinstance(value, list):
                    values = list(filter(bool, [v.strip() for v in value]))
                else:
                    values = [v.strip() for v in value.split(',')]
                value = values if len(values) > 1 else values[0]

            metadata[key] = value
        return metadata

    def process_entry(self, context: Context, entry_uri: EntryURI, changed_files: set[Path] | None = None) -> None:
        if entry_uri.lower().endswith('.md'):
            if config.fast_rebuilds and changed_files and (config.content_path / entry_uri) not in changed_files:
                return

            self.markdown.context = context
            markdown_text = (config.content_path / entry_uri).read_text()
            html = self.markdown.reset().convert(markdown_text)

            context['entries'][entry_uri].update({
                **self.parse_metadata(self.markdown.Meta),
                'body': html,
                'table_of_contents': self.markdown.toc_tokens,
                'url': f"{config.site_url}/{str(Path(entry_uri).with_suffix(config.html_url_extension))}",
            })



from pathlib import Path
from ursus.config import config
from ursus.linters import RegexLinter, LinterResult
import logging
import re


class OrphanFootnotesLinter(RegexLinter):
    """
    Count footnotes; make sure that they all appear at least twice
    """
    file_suffixes = ('.md', )
    regex = re.compile(r'(?P<footnote>\[\^(?P<id>\d+)\])(?P<colon>:?)')

    def lint(self, file_path: Path) -> LinterResult:
        if self.file_suffixes and file_path.suffix.lower() not in self.file_suffixes:
            return

        footnotes: dict[str, list[dict]] = {}

        with (config.content_path / file_path).open() as file:
            for line_no, line in enumerate(file.readlines()):
                for match in self.regex.finditer(line):
                    footnotes.setdefault(match['id'], [])
                    footnotes[match['id']].append({
                        'position': (line_no, *match.span()),
                        'is_definition': bool(match['colon']),
                    })

        for footnote_id, occurences in footnotes.items():
            definitions = [o for o in occurences if o['is_definition']]

            if len(definitions) == 0:
                for occurence in occurences:
                    yield occurence['position'], f"Undefined footnote: [^{footnote_id}]", logging.ERROR
            elif len(definitions) > 1:
                for definition in definitions:
                    yield definition['position'], f"Multiple footnote definitions: [^{footnote_id}]", logging.ERROR
            elif len(definitions) == len(occurences):
                for occurence in occurences:
                    yield occurence['position'], f"Unused footnote: [^{footnote_id}]", logging.ERROR



from ursus.context_processors.get_entries import get_entries


entries = {
    'blog/hello.md': {
        'title': 'Hello world',
        'month': 'January',
        'short_title': 'Hello',
        'description': 'First post on this blog!',
    },
    'blog/bonjour.md': {
        'title': 'Bonjour monde',
        'month': 'October',
        'short_title': 'Bonjour',
        'description': 'Deuxième post sur ce blogue',
    },
    'blog/hallo.md': {
        'title': 'Hallo Welt',
        'month': 'September',
        'short_title': 'Hallo',
        'description': 'Dritter Beitrag auf diesem Blog',
    },
    'recipes/pancakes.md': {
        'title': 'Pancake recipe',
        'description': 'Flat and delicious',
        'ingredients': ['Flour', 'eggs', 'milk'],
    },
    'home.md': {
        'title': 'Home page',
        'content': 'Boy, what a home page!',
    },
    'empty.md': {
    },
}


def test_sort_by_function():
    def month_getter(entry):
        return [
            'January',
            'February',
            'March',
            'April',
            'May',
            'June',
            'July',
            'August',
            'September',
            'October',
            'November',
            'December',
        ].index(entry['month'])

    assert get_entries(entries, 'blog', sort_by=month_getter) == [
        entries['blog/hello.md'],
        entries['blog/hallo.md'],
        entries['blog/bonjour.md'],
    ]


def test_sort_by_one_key():
    assert get_entries(entries, 'blog', sort_by='title') == [
        entries['blog/bonjour.md'],
        entries['blog/hallo.md'],
        entries['blog/hello.md'],
    ]


def test_sort_by_multiple_keys():
    assert get_entries(entries, 'blog', sort_by=['short_title', 'title']) == [
        entries['blog/bonjour.md'],
        entries['blog/hallo.md'],
        entries['blog/hello.md'],
    ]


def test_filter_by():
    def uri_length_filter(entry_uri, entry):
        return len(entry_uri) <= 13

    filtered_entries = get_entries(entries, None, filter_by=uri_length_filter)
    assert len(filtered_entries) == 4
    assert entries['blog/hallo.md'] in filtered_entries
    assert entries['blog/hello.md'] in filtered_entries
    assert entries['home.md'] in filtered_entries
    assert entries['empty.md'] in filtered_entries


def test_no_namespace():
    assert len(get_entries(entries)) == 6


def test_namespace():
    assert get_entries(entries, 'recipes') == [
        entries['recipes/pancakes.md'],
    ]



from pathlib import Path
from ursus.config import config
from ursus.linters import Linter, LinterResult
from ursus.linters.markdown import MarkdownLinksLinter
from ursus.utils import is_image, get_files_in_path
import logging


class UnusedImagesLinter(Linter):
    """
    Verify that internal links point to existing entries. If the URL has a fragment,
    it should point to an existing title fragment.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_entry_links = set()

        link_regex = MarkdownLinksLinter.regex

        for markdown_path in get_files_in_path(config.content_path, suffix='.md'):
            with (config.content_path / markdown_path).open() as file:
                for line_no, line in enumerate(file.readlines()):
                    for match in link_regex.finditer(line):
                        entry_uri = match['url'].removeprefix('/')
                        self.all_entry_links.add(entry_uri)

    def lint(self, file_path: Path) -> LinterResult:
        if is_image(config.content_path / file_path) and str(file_path) not in self.all_entry_links:
            yield None, "Unused image", logging.WARNING



from importlib import import_module
from markdown.extensions.meta import BEGIN_RE, META_RE, META_MORE_RE, END_RE
from pathlib import Path
from PIL import Image, ImageCms
from PIL.Image import Image as ImageType
from types import ModuleType
from typing import Any, Iterator, Tuple, List
from ursus.config import config
from ursus.context_processors import Context, EntryURI
from xml.etree import ElementTree
import fitz
import imagesize
import io
import logging
import os
import shutil
import sys


def log_color(level: int = logging.INFO) -> str:
    if os.getenv('TERM') not in ['xterm-color', 'xterm-256color', 'screen']:
        return ''

    return {
        logging.DEBUG: '',
        logging.INFO: '\033[37m\033[0;100m',
        logging.WARNING: '\033[1;90m\033[43m',
        logging.ERROR: '\033[37m\033[41m',
        logging.CRITICAL: '\033[37m\033[41m',
    }[level]


def log_color_end() -> str:
    if os.getenv('TERM') not in ['xterm-color', 'xterm-256color', 'screen']:
        return ''
    return '\033[0m'


def import_class(import_path):
    module_name, class_name = import_path.rsplit('.', 1)
    module = import_module(module_name)
    return getattr(module, class_name)


def import_module_or_path(module_or_path: str | Path) -> ModuleType:
    """
    Imports a module (path.to.module.class or path/to/module.py)

    Args:
        module_or_path (str): path.to.module, or path/to/module.py
    """
    file_path = Path(module_or_path)
    if file_path.exists():
        sys.path.append(str(file_path.parent))
        return import_module(file_path.with_suffix('').name)
    return import_module(str(module_or_path))


def is_ignored_file(path: Path, root_path: Path) -> bool:
    """Returns whether a file should be ignored by ursus.
    Args:
        path (Path): Path to the file
        root_path (Path): Root path. The name of parent directories above root_path are not considered.

    Returns:
        bool: True if the file or any of its parents (up to root_path) start with _ or .
    """
    return (
        path.stem.startswith(('_', '.'))
        or any([
            p.stem.startswith(('_', '.'))
            for p in path.relative_to(root_path).parents
        ])
    )


def is_image(path: Path) -> bool:
    """Whether the given path points to an image

    Args:
        path (Path): The path to the file

    Returns:
        bool: Whether this file is an image
    """
    assert path.is_absolute(), 'is_image must be called with an absolute path'

    # Notably missing: .heif, .heic
    image_suffixes = ('.apng', '.avif', '.gif', '.jpg', '.jpeg', '.png', '.svg', '.webp')
    return path.is_file() and path.suffix.lower() in image_suffixes


def is_pdf(path: Path) -> bool:
    """Whether the given path points to a PDF

    Args:
        path (Path): The path to the file

    Returns:
        bool: Whether this file is a PDF
    """
    assert path.is_absolute(), 'is_pdf must be called with an absolute path'
    return path.is_file() and path.suffix.lower() == '.pdf'


def is_svg(path: Path) -> bool:
    """Whether the given path points to an SVG image

    Args:
        path (Path): The path to the file

    Returns:
        bool: Whether this file is an SVG image
    """
    assert path.is_absolute(), 'is_svg must be called with an absolute path'
    return path.is_file() and path.suffix.lower() == '.svg'


def is_raster_image(path: Path) -> bool:
    return is_image(path) and not is_svg(path)


def get_image_size(path: Path) -> tuple[int, int]:
    """
    Args:
        path (Path): The absolute Path to an image
    Returns:
        tuple: The image width and height as a tuple
    """
    try:
        width, height = imagesize.get(path)
        if width == height == -1:
            logging.debug(f"Can't get image size, trying again with Pillow: {path}")
            with Image.open(path) as pil_image:
                width, height = pil_image.size
        return width, height
    except Exception as e:
        raise Exception(f"Invalid image size: {path}") from e


def get_files_in_path(path: Path, whitelist: set[Path] | None = None, suffix: str | None = None) -> list[Path]:
    """
    Returns a list of valid, visible files under a given path. The returned paths are relative to the supplied path.

    Args:
        path (Path): The path under which to find files
        whitelist (set, optional): Only include files that are part of this whitelist
        suffix (str, optional): Only include files with this suffix
    Returns:
        list[Path]: A list of files in this path
    """
    if whitelist:
        files = []
        for f in whitelist:
            if (not f.is_absolute()) and (path / f).exists():
                files.append(path / f)
            elif f.is_absolute() and f.is_relative_to(path):
                files.append(f)
    else:
        files = list(path.rglob('[!._]*' + (suffix or '')))

    return [
        f.relative_to(path) for f in files
        if (
            f.is_file()
            and not is_ignored_file(f, path)
            and suffix in (None, f.suffix)
        )
    ]


def copy_file(input_path: Path, output_path: Path) -> None:
    """Copies a file

    Args:
        input_path (Path): The absolute path of the file to copy
        output_path (Path): The absolute path of the file destination
    """
    assert input_path.is_absolute(), f"input_path {str(input_path)} is relative. It must be absolute."
    assert output_path.is_absolute(), f"output_path {str(output_path)} is relative. It must be absolute."

    output_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(input_path, output_path)


def convert_to_srgb(pil_image: ImageType) -> ImageType:
    """
    Convert PIL image to sRGB color space (if possible), since the profile
    information is stripped when resizing the images.
    """

    icc = pil_image.info.get('icc_profile', '')
    if icc:
        io_handle = io.BytesIO(icc)     # virtual file
        src_profile = ImageCms.ImageCmsProfile(io_handle)
        dst_profile = ImageCms.createProfile('sRGB')
        try:
            pil_image = ImageCms.profileToProfile(pil_image, src_profile, dst_profile) or pil_image
        except ImageCms.PyCMSError:
            logging.warning("Could not convert image color profile. Skipping conversion.")
    return pil_image


def make_image_thumbnail(pil_image: ImageType, max_size, output_path: Path) -> None:
    """Creates a thumbnail of an image. Strips EXIF metadata.

    Args:
        pil_image (Image): A Pillow Image object containing the image to resize
        max_size (TYPE): Max width and height of the preview image
        output_path (Path): Path to the resulting preview
    """
    assert output_path.is_absolute(), f"output_path {str(output_path)} is relative. It must be absolute."

    pil_image = convert_to_srgb(pil_image)
    pil_image.thumbnail(max_size, Image.Resampling.LANCZOS)
    save_args = {'optimize': True}
    if output_path.suffix.lower() == '.jpg':
        save_args['progressive'] = True
    elif output_path.suffix.lower() == '.webp':
        save_args['exact'] = True

    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Note: The saved image is stripped of EXIF data
    pil_image.save(output_path, **save_args)


def make_pdf_thumbnail(pdf_path: Path, max_size, output_path: Path) -> None:
    """Creates an image preview of a PDF file

    Args:
        pdf_path (Path): Path to the PDF file to preview
        max_size (TYPE): Max width and height of the preview image
        output_path (Path): Path to the resulting preview
    """
    assert output_path.is_absolute(), f"output_path {str(output_path)} is relative. It must be absolute."

    width, height = max_size
    doc = fitz.open(pdf_path)
    pixmap = doc[0].get_pixmap(alpha=False)
    thumbnail = Image.frombytes('RGB', [pixmap.width, pixmap.height], pixmap.samples)
    make_image_thumbnail(thumbnail, max_size, output_path)


def get_image_transforms(original_path: Path) -> Iterator[dict]:
    """
    Yields a list of image transforms that apply to a file.

    Args:
        original_path (Path): Path of the original file/image to transform, relative to config.content_path

    Yields:
        Iterator[dict]: A list of image transforms that apply to this file.
    """
    suffix_to_mimetype = {
        '.apng': 'image/apng',
        '.avif': 'image/avif',
        '.gif': 'image/avif',
        '.jpg': 'image/jpeg',
        '.jpeg': 'image/jpeg',
        '.pdf': 'application/pdf',
        '.png': 'image/png',
        '.svg': 'image/svg+xml',
        '.webp': 'image/webp',
    }

    assert not original_path.is_absolute(), "original_path must be a relative path"

    for key, transform in config.image_transforms.items():
        includes = transform.get('include', ['*'])
        includes = [includes] if isinstance(includes, str) else includes

        excludes = transform.get('exclude', [])
        excludes = [excludes] if isinstance(excludes, str) else excludes

        transform_applies_to_file = (
            any(original_path.match(pattern) for pattern in includes)
            and not any(original_path.match(pattern) for pattern in excludes)
        )

        if transform_applies_to_file:
            # Normalise and deduplicate suffixes
            # For orig_image.JPG, ('original', 'jpg') becomes ('.jpg')
            output_suffixes = set([
                original_path.suffix.lower() if t == 'original' else '.' + t
                for t in transform.get('output_types', ['original'])
            ])

            for suffix in output_suffixes:
                if suffix == original_path.suffix.lower():
                    output_image_path = original_path
                else:
                    if is_svg(config.content_path / original_path):
                        # .svg images are not be converted
                        continue

                    output_image_path = original_path.with_suffix(suffix)

                yield {
                    **transform,
                    'is_default': key == '',
                    'input_mimetype': suffix_to_mimetype[original_path.suffix.lower()],
                    'output_mimetype': suffix_to_mimetype[output_image_path.suffix.lower()],
                    'output_path': output_image_path.parent / key / output_image_path.name,  # It works if key is empty
                }


def make_picture_element(context: Context, entry_uri: EntryURI, img_attrs={}, a_attrs=None, sizes=None) -> ElementTree.Element:
    """
    Creates a responsive HTML <picture> element
    """
    # Mimetypes with broad support that don't require their own <source> element
    standard_mimetypes = ('image/png', 'image/jpeg', 'image/svg+xml')

    # TODO: SVG? PDF?
    default_src = None

    # Build a list of srcsets grouped by mimetype
    sources_by_mimetype: dict[str, list[str]] = {}
    for transform in context['entries'][entry_uri]['transforms']:
        width = transform['max_size'][0]
        mimetype = transform['output_mimetype']

        if mimetype.startswith('image/'):
            srcset_part = f"{config.site_url}/{str(transform['output_path'])} {width}w"
            sources_by_mimetype.setdefault(mimetype, [])
            sources_by_mimetype[mimetype].append(srcset_part)

            if not default_src and mimetype in standard_mimetypes:
                default_src = transform['output_path']

    # Create a <picture> with <source type="" srcset=""> for each mimetype
    picture = ElementTree.Element('picture')
    for mimetype, srcset_elements in sources_by_mimetype.items():
        source = ElementTree.SubElement(picture, 'source', attrib={
            'type': mimetype,
            'srcset': ", ".join(srcset_elements)
        })
        if sizes:
            source.attrib['sizes'] = sizes

    # Add an <img> with the default image to the <picture>
    img = ElementTree.SubElement(picture, 'img', attrib=img_attrs)
    assert default_src is not None, f"default_src is None for {entry_uri}"
    if 'width' in context['entries'][entry_uri]:
        img.attrib['width'] = str(context['entries'][entry_uri]['width'])
    if 'height' in context['entries'][entry_uri]:
        img.attrib['height'] = str(context['entries'][entry_uri]['height'])
    img.attrib['loading'] = 'lazy'
    img.attrib['src'] = f"{config.site_url}/{str(default_src)}"

    return picture


def make_figure_element(context: Context, entry_uri: EntryURI, img_attrs={}, a_attrs=None, sizes=None):
    """
    Creates a responsive HTML <figure> element with the image title as <figcaption>. Returns a simple <picture> if there
    is no title.
    """
    image = make_picture_element(context, entry_uri, img_attrs, sizes=sizes)
    if a_attrs and a_attrs.get('href'):
        a_attrs['target'] = '_blank'
        wrapped_image = ElementTree.Element('a', a_attrs)
        wrapped_image.append(image)
    else:
        wrapped_image = image

    if not img_attrs.get('title'):
        return wrapped_image

    figure = ElementTree.Element('figure')
    figure.append(wrapped_image)

    figcaption = ElementTree.SubElement(figure, 'figcaption')
    figcaption.text = img_attrs['title']

    return figure


def parse_markdown_head_matter(lines: list[str]) -> Tuple[dict[str, List[Any]], dict[str, Tuple[int, int, int]]]:
    """
    Turns markdown head matter into a dictionary. Returns the dictionary, and the position of each dictionary key in
    the file (to allow linters to highlight invalid head matter keys)
    """
    meta: dict[str, List[Any]] = {}
    field_positions: dict[str, Tuple[int, int, int]] = {}

    if lines and BEGIN_RE.match(lines[0]):
        lines.pop(0)

    # The key is a string. The value is always a list, although it usually has only one item
    # This mirrors the behaviour of Python-Markdown's meta extension, which Ursus relies on.
    for line_no, line in enumerate(lines):
        m1 = META_RE.match(line)
        if line.strip() == '' or END_RE.match(line):
            break  # blank line or end of YAML header - done
        if m1:
            key = m1.group('key').lower().strip()
            value = m1.group('value').strip()
            try:
                meta[key].append(value)
            except KeyError:
                meta[key] = [value]
            field_positions[key] = (line_no + 1, 0, len(line) - 1)
        else:
            m2 = META_MORE_RE.match(line)
            if m2 and key:
                # Add another line to existing key
                meta[key].append(m2.group('value').strip())
            else:
                lines.insert(0, line)
                break
    return meta, field_positions


def format_markdown_head_matter(metadata: dict[str, List[int]]) -> str:
    lines = []

    for key, value_list in metadata.items():
        lines.append(f'{key.lower()}: {value_list[0]}')
        lines.extend([f'    {value}' for value in value_list[1:]])

    return '\n'.join(['---', *lines, '---'])



from . import Renderer
from lunr import lunr
from pathlib import Path
from ursus.config import config
from ursus.context_processors import Context, Entry, EntryURI
import json
import logging


logger = logging.getLogger(__name__)


class LunrIndexRenderer(Renderer):
    """
    Renders a .json search index for Lunr.js. The resulting file is a dictionary
    with two keys:

    - 'index': An index object that can be passed to Lunr.js
    - 'documents': A dict of entry URI to documents that can be used to render
        search results (titles, URLs, excerpts, etc.)
    """

    def __init__(self):
        super().__init__()

    def get_index_for_entry(self, index_config: dict, entry_uri: EntryURI, entry: Entry):
        if Path(entry_uri).match(index_config['uri_pattern']):
            # Data used to build the Lunr.js index (indexed fields, document boost)
            indexed_document = (
                {
                    **{field: entry.get(field, '') for field in config.lunr_indexes['indexed_fields']},
                },
                {
                    'boost': index_config.get('boost', 1)
                }
            )

            # Metadata about the documents, used to render the search box (title, URL, etc)
            returned_document = {
                field: entry.get(field) for field in index_config['returned_fields']
            }

            yield indexed_document, returned_document

    def render(self, context: Context, changed_files: set[Path] | None = None) -> set[Path]:
        if config.fast_rebuilds:
            return set()

        index_output_path = config.output_path / config.lunr_index_output_path
        logger.info(f"Generating search index at {config.lunr_index_output_path}")

        indexed_documents = []
        returned_documents = {}

        document_ref = 0
        for index_config in config.lunr_indexes.get('indexes', []):
            for entry_uri, entry in context['entries'].items():
                for indexed_document, returned_document in self.get_index_for_entry(index_config, entry_uri, entry):
                    # indexed_document contains the fields that are included in the Lunr index
                    # returned_document contains information about the entry (title, URL)
                    # The ref attribute connects a Lunr search result to a returned_document
                    indexed_document[0]['ref'] = document_ref
                    indexed_documents.append(indexed_document)
                    returned_documents[document_ref] = returned_document
                    document_ref += 1

        index = lunr(
            ref='ref',
            fields=config.lunr_indexes.get('indexed_fields', []),
            documents=indexed_documents
        )

        index_output_path.parent.mkdir(parents=True, exist_ok=True)
        with index_output_path.open('w+') as index_file:
            json.dump({
                'index': index.serialize(),
                'documents': returned_documents,
            }, index_file)

        return set([config.lunr_index_output_path])



from setuptools import setup, find_packages
from pathlib import Path

root_dir = Path(__file__).parent

long_description = (root_dir / "README.md").read_text()
requirements_path = root_dir / "requirements.txt"
install_requires = requirements_path.read_text().strip().split('\n')

setup(
    name='ursus_ssg',
    version='1.4.2',
    description='Static site generator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://github.com/all-about-berlin/ursus',
    author='Nicolas Bouliane',
    author_email='contact@nicolasbouliane.com',
    license='MIT',
    packages=find_packages(),
    scripts=['bin/ursus'],
    entry_points={
        'markdown.extensions': [
            'better_footnotes = ursus.context_processors.markdown:FootnotesExtension',
            'jinja = ursus.context_processors.markdown:JinjaExtension',
            'responsive_images = ursus.context_processors.markdown:ResponsiveImagesExtension',
            'superscript = ursus.context_processors.markdown:SuperscriptExtension',
            'tasklist = ursus.context_processors.markdown:TaskListExtension',
        ]
    },
    package_data={
        'ursus.babel': ['*', ],
        'ursus': ['py.typed', ],
    },
    python_requires='>=3.11',
    install_requires=install_requires,
    zip_safe=False,
)
